@@ -50,6 +50,7 @@
 #include <gdbus.h>
 
 #include "glib-helper.h"
+#include "../src/manager.h"
 #include "../src/adapter.h"
 #include "../src/device.h"
 
@@ -102,6 +103,7 @@ static GSList *adapters = NULL;
 static GSList *devices = NULL;
 
 static struct enabled_interfaces enabled = {
+	.hfp		= TRUE,
 	.headset	= TRUE,
 	.gateway	= FALSE,
 	.sink		= TRUE,
@@ -123,118 +125,69 @@ static struct audio_adapter *find_adapter(GSList *list, const char *path)
 	return NULL;
 }
 
-static struct audio_adapter *find_adapter_by_address(GSList *list,
-						const bdaddr_t *src)
-{
-	GSList *l;
-
-	for (l = list; l; l = l->next) {
-		struct audio_adapter *adapter = l->data;
-
-		if (bacmp(&adapter->src, src) == 0)
-			return adapter;
-	}
-
-	return NULL;
-}
-
-static uint16_t get_service_uuid(const sdp_record_t *record)
-{
-	sdp_list_t *classes;
-	uuid_t uuid;
-	uint16_t uuid16 = 0;
-
-	if (sdp_get_service_classes(record, &classes) < 0) {
-		error("Unable to get service classes from record");
-		return 0;
-	}
-
-	memcpy(&uuid, classes->data, sizeof(uuid));
-
-	if (!sdp_uuid128_to_uuid(&uuid)) {
-		error("Not a 16 bit UUID");
-		sdp_list_free(classes, free);
-		return 0;
-	}
-
-	if (uuid.type == SDP_UUID32) {
-		if (uuid.value.uuid32 > 0xFFFF) {
-			error("Not a 16 bit UUID");
-			goto done;
-		}
-		uuid16 = (uint16_t) uuid.value.uuid32;
-	} else
-		uuid16 = uuid.value.uuid16;
-
-done:
-	sdp_list_free(classes, free);
-
-	return uuid16;
-}
-
 gboolean server_is_enabled(bdaddr_t *src, uint16_t svc)
 {
-	struct audio_adapter *adp;
-	gboolean ret;
-
-	adp = find_adapter_by_address(adapters, src);
-	if (!adp)
-		return FALSE;
-
 	switch (svc) {
 	case HEADSET_SVCLASS_ID:
-		ret = (adp->hsp_ag_server != NULL);
-		break;
+		return enabled.headset;
 	case HEADSET_AGW_SVCLASS_ID:
-		ret = (adp->hsp_hs_server != NULL);
-		break;
+		return enabled.gateway;
 	case HANDSFREE_SVCLASS_ID:
-		ret = (adp->hfp_ag_server != NULL);
-		break;
+		return enabled.headset && enabled.hfp;
 	case HANDSFREE_AGW_SVCLASS_ID:
-		ret = FALSE;
-		break;
+		return  FALSE;
 	case AUDIO_SINK_SVCLASS_ID:
 		return enabled.sink;
 	case AV_REMOTE_TARGET_SVCLASS_ID:
 	case AV_REMOTE_SVCLASS_ID:
 		return enabled.control;
 	default:
-		ret = FALSE;
-		break;
+		return FALSE;
 	}
-
-	return ret;
 }
 
-static void handle_record(sdp_record_t *record, struct audio_device *device)
+static void handle_uuid(const char *uuidstr, struct audio_device *device)
 {
+	uuid_t uuid;
 	uint16_t uuid16;
 
-	uuid16 = get_service_uuid(record);
-
-	if (!server_is_enabled(&device->src, uuid16))
+	if (bt_string2uuid(&uuid, uuidstr) < 0) {
+		error("%s not detected as an UUID-128", uuidstr);
 		return;
+	}
+
+	if (!sdp_uuid128_to_uuid(&uuid) && uuid.type != SDP_UUID16) {
+		error("Could not convert %s to a UUID-16", uuidstr);
+		return;
+	}
+
+	uuid16 = uuid.value.uuid16;
+
+	if (!server_is_enabled(&device->src, uuid16)) {
+		debug("audio handle_uuid: server not enabled for %s (0x%04x)",
+				uuidstr, uuid16);
+		return;
+	}
 
 	switch (uuid16) {
 	case HEADSET_SVCLASS_ID:
 		debug("Found Headset record");
 		if (device->headset)
-			headset_update(device, record, uuid16);
+			headset_update(device, uuid16, uuidstr);
 		else
-			device->headset = headset_init(device,
-							record, uuid16);
+			device->headset = headset_init(device, uuid16,
+							uuidstr);
 		break;
 	case HEADSET_AGW_SVCLASS_ID:
 		debug("Found Headset AG record");
 		break;
 	case HANDSFREE_SVCLASS_ID:
-		debug("Found Hansfree record");
+		debug("Found Handsfree record");
 		if (device->headset)
-			headset_update(device, record, uuid16);
+			headset_update(device, uuid16, uuidstr);
 		else
-			device->headset = headset_init(device,
-							record, uuid16);
+			device->headset = headset_init(device, uuid16,
+							uuidstr);
 		break;
 	case HANDSFREE_AGW_SVCLASS_ID:
 		debug("Found Handsfree AG record");
@@ -473,7 +426,8 @@ static void ag_io_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 			const bdaddr_t *dst, gpointer data)
 {
 	struct audio_adapter *adapter = data;
-	const char *uuid;
+	const char *server_uuid, *remote_uuid;
+	uint16_t svclass;
 	struct audio_device *device;
 	gboolean hfp_active;
 
@@ -484,16 +438,23 @@ static void ag_io_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	if (chan == adapter->hsp_ag_server) {
 		hfp_active = FALSE;
-		uuid = HSP_AG_UUID;
+		server_uuid = HSP_AG_UUID;
+		remote_uuid = HSP_HS_UUID;
+		svclass = HEADSET_SVCLASS_ID;
 	} else {
 		hfp_active = TRUE;
-		uuid = HFP_AG_UUID;
+		server_uuid = HFP_AG_UUID;
+		remote_uuid = HFP_HS_UUID;
+		svclass = HANDSFREE_SVCLASS_ID;
 	}
 
-	device = manager_find_device(dst, NULL, FALSE);
+	device = manager_get_device(src, dst);
 	if (!device)
 		goto drop;
 
+	if (!device->headset)
+		btd_device_add_uuid(device->btd_dev, remote_uuid);
+
 	if (headset_get_state(device) > HEADSET_STATE_DISCONNECTED) {
 		debug("Refusing new connection since one already exists");
 		goto drop;
@@ -508,8 +469,8 @@ static void ag_io_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	headset_set_state(device, HEADSET_STATE_CONNECT_IN_PROGRESS);
 
-	err = btd_request_authorization(&device->src, &device->dst, uuid,
-				auth_cb, device);
+	err = btd_request_authorization(&device->src, &device->dst,
+					server_uuid, auth_cb, device);
 	if (err < 0) {
 		debug("Authorization denied: %s", strerror(-err));
 		headset_set_state(device, HEADSET_STATE_DISCONNECTED);
@@ -534,7 +495,7 @@ static int headset_server_init(struct audio_adapter *adapter)
 {
 	uint8_t chan = DEFAULT_HS_AG_CHANNEL;
 	sdp_record_t *record;
-	gboolean hfp = TRUE, master = TRUE;
+	gboolean master = TRUE;
 	GError *err = NULL;
 	uint32_t features, flags;
 
@@ -549,15 +510,6 @@ static int headset_server_init(struct audio_adapter *adapter)
 			err = NULL;
 		} else
 			master = tmp;
-
-		tmp = g_key_file_get_boolean(config, "Headset", "HFP",
-						&err);
-		if (err) {
-			debug("audio.conf: %s", err->message);
-			g_error_free(err);
-			err = NULL;
-		} else
-			hfp = tmp;
 	}
 
 	flags = RFCOMM_LM_AUTH | RFCOMM_LM_ENCRYPT;
@@ -585,7 +537,7 @@ static int headset_server_init(struct audio_adapter *adapter)
 
 	features = headset_config_init(config);
 
-	if (!hfp)
+	if (!enabled.hfp)
 		return 0;
 
 	chan = DEFAULT_HF_AG_CHANNEL;
@@ -674,25 +626,22 @@ static int gateway_server_init(struct audio_adapter *adapter)
 	return 0;
 }
 
-static int audio_probe(struct btd_device *device, GSList *records)
+static int audio_probe(struct btd_device *device, GSList *uuids)
 {
 	struct btd_adapter *adapter = device_get_adapter(device);
-	const gchar *path = device_get_path(device);
 	bdaddr_t src, dst;
-	struct audio_device *dev;
+	struct audio_device *audio_dev;
 
 	adapter_get_address(adapter, &src);
 	device_get_address(device, &dst);
 
-	dev = manager_find_device(&dst, NULL, FALSE);
-	if (!dev) {
-		dev = device_register(connection, path, &src, &dst);
-		if (!dev)
-			return -EINVAL;
-		devices = g_slist_append(devices, dev);
+	audio_dev = manager_get_device(&src, &dst);
+	if (!audio_dev) {
+		debug("audio_probe: unable to get a device object");
+		return -1;
 	}
 
-	g_slist_foreach(records, (GFunc) handle_record, dev);
+	g_slist_foreach(uuids, (GFunc) handle_uuid, audio_dev);
 
 	return 0;
 }
@@ -918,6 +867,8 @@ int audio_manager_init(DBusConnection *conn, GKeyFile *conf)
 {
 	char **list;
 	int i;
+	gboolean b;
+	GError *err;
 
 	connection = dbus_connection_ref(conn);
 
@@ -958,6 +909,16 @@ int audio_manager_init(DBusConnection *conn, GKeyFile *conf)
 	}
 	g_strfreev(list);
 
+	err = NULL;
+	b = g_key_file_get_boolean(config, "Headset", "HFP",
+					&err);
+	if (err) {
+		debug("audio.conf: %s", err->message);
+		g_error_free(err);
+		err = NULL;
+	} else
+		enabled.hfp = b;
+
 proceed:
 	if (enabled.headset) {
 		telephony_init();
@@ -1057,3 +1018,47 @@ struct audio_device *manager_find_device(const bdaddr_t *bda, const char *interf
 
 	return NULL;
 }
+
+struct audio_device *manager_get_device(const bdaddr_t *src,
+					const bdaddr_t *dst)
+{
+	struct audio_device *dev;
+	struct btd_adapter *adapter;
+	struct btd_device *device;
+	char addr[18];
+	const char *path;
+
+	dev = manager_find_device(dst, NULL, FALSE);
+	if (dev)
+		return dev;
+
+
+	ba2str(src, addr);
+
+	adapter = manager_find_adapter(src);
+	if (!adapter) {
+		error("Unable to get a btd_adapter object for %s",
+				addr);
+		return NULL;
+	}
+
+	ba2str(dst, addr);
+
+	device = adapter_get_device(connection, adapter, addr);
+	if (!device) {
+		error("Unable to get btd_device object for %s", addr);
+		return NULL;
+	}
+
+	path = device_get_path(device);
+
+	dev = device_register(connection, path, src, dst);
+	if (!dev)
+		return NULL;
+
+	dev->btd_dev = device;
+
+	devices = g_slist_append(devices, dev);
+
+	return dev;
+}
