@@ -806,6 +806,9 @@ static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
 	struct btd_adapter *adapter = data;
 	const char *path;
 
+	if (adapter->discov_timeout == timeout && timeout == 0)
+		return dbus_message_new_method_return(msg);
+
 	if (adapter->discov_timeout_id) {
 		g_source_remove(adapter->discov_timeout_id);
 		adapter->discov_timeout_id = 0;
@@ -820,10 +823,9 @@ static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
 
 	path = dbus_message_get_path(msg);
 
-	dbus_connection_emit_property_changed(conn, path,
-					ADAPTER_INTERFACE,
-					"DiscoverableTimeout",
-					DBUS_TYPE_UINT32, &timeout);
+	emit_property_changed(conn, path,
+				ADAPTER_INTERFACE, "DiscoverableTimeout",
+				DBUS_TYPE_UINT32, &timeout);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -899,10 +901,8 @@ static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
 
 	path = dbus_message_get_path(msg);
 
-	dbus_connection_emit_property_changed(conn, path,
-					ADAPTER_INTERFACE,
-					"Name", DBUS_TYPE_STRING,
-					&name);
+	emit_property_changed(conn, path, ADAPTER_INTERFACE, "Name",
+				DBUS_TYPE_STRING, &name);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -1030,9 +1030,8 @@ static DBusMessage *remove_bonding(DBusConnection *conn, DBusMessage *msg,
 
 		const gchar *dev_path = device_get_path(device);
 
-		dbus_connection_emit_property_changed(conn, dev_path,
-					DEVICE_INTERFACE, "Paired",
-					DBUS_TYPE_BOOLEAN, &paired);
+		emit_property_changed(conn, dev_path, DEVICE_INTERFACE,
+					"Paired", DBUS_TYPE_BOOLEAN, &paired);
 	}
 
 proceed:
@@ -1579,38 +1578,33 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	/* Address */
 	property = srcaddr;
-	dbus_message_iter_append_dict_entry(&dict, "Address",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Address", DBUS_TYPE_STRING, &property);
 
 	/* Name */
 	memset(str, 0, sizeof(str));
 	strncpy(str, (char *) adapter->dev.name, 248);
 	property = str;
 
-	dbus_message_iter_append_dict_entry(&dict, "Name",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &property);
 
 	/* Mode */
 	property = mode2str(adapter->mode);
 
-	dbus_message_iter_append_dict_entry(&dict, "Mode",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Mode", DBUS_TYPE_STRING, &property);
 
 	/* Powered */
 	if (main_opts.offmode == HCID_OFFMODE_DEVDOWN)
 		value = adapter->up ? TRUE : FALSE;
 	else
 		value = adapter->scan_mode == SCAN_DISABLED ? FALSE : TRUE;
-	dbus_message_iter_append_dict_entry(&dict, "Powered",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Powered", DBUS_TYPE_BOOLEAN, &value);
 
 	/* Discoverable */
 	value = adapter->scan_mode & SCAN_INQUIRY ? TRUE : FALSE;
-	dbus_message_iter_append_dict_entry(&dict, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Discoverable", DBUS_TYPE_BOOLEAN, &value);
 
 	/* DiscoverableTimeout */
-	dbus_message_iter_append_dict_entry(&dict, "DiscoverableTimeout",
+	dict_append_entry(&dict, "DiscoverableTimeout",
 				DBUS_TYPE_UINT32, &adapter->discov_timeout);
 
 	if (adapter->state & PERIODIC_INQUIRY || adapter->state & STD_INQUIRY)
@@ -1619,8 +1613,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		value = FALSE;
 
 	/* Discovering */
-	dbus_message_iter_append_dict_entry(&dict, "Discovering",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Discovering", DBUS_TYPE_BOOLEAN, &value);
 
 	/* Devices */
 	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
@@ -1628,8 +1621,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		struct btd_device *dev = l->data;
 		devices[i] = (char *) device_get_path(dev);
 	}
-	dbus_message_iter_append_dict_entry(&dict, "Devices",
-			DBUS_TYPE_ARRAY, &devices);
+	dict_append_array(&dict, "Devices", DBUS_TYPE_OBJECT_PATH, &devices, i);
 	g_free(devices);
 
 	dbus_message_iter_close_container(&iter, &dict);
@@ -1898,7 +1890,7 @@ static DBusMessage *create_device(DBusConnection *conn,
 
 	device_set_temporary(device, FALSE);
 
-	device_browse(device, conn, msg, NULL);
+	device_browse(device, conn, msg, NULL, FALSE);
 
 	adapter->devices = g_slist_append(adapter->devices, device);
 
@@ -2271,7 +2263,6 @@ static void create_stored_device_from_profiles(char *key, char *value,
 	struct btd_adapter *adapter = user_data;
 	GSList *uuids = bt_string2list(value);
 	struct btd_device *device;
-	sdp_list_t *records;
 	bdaddr_t dst;
 	char srcaddr[18], dstaddr[18];
 
@@ -2291,13 +2282,9 @@ static void create_stored_device_from_profiles(char *key, char *value,
 	device_get_address(device, &dst);
 	ba2str(&dst, dstaddr);
 
-	records = read_records(srcaddr, dstaddr);
-
-	device_probe_drivers(device, uuids, records);
-
-	if (records)
-		sdp_list_free(records, (sdp_free_func_t) sdp_record_free);
+	device_probe_drivers(device, uuids);
 
+	g_slist_foreach(uuids, (GFunc) g_free, NULL);
 	g_slist_free(uuids);
 }
 
@@ -2443,22 +2430,20 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 
 	pmode = mode2str(adapter->mode);
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Mode",
-					DBUS_TYPE_STRING, &pmode);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Mode", DBUS_TYPE_STRING, &pmode);
 
 	powered = adapter->scan_mode == SCAN_DISABLED ? FALSE : TRUE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Powered", DBUS_TYPE_BOOLEAN, &powered);
 
 	discoverable = adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY) ? TRUE
 				: FALSE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &discoverable);
+	emit_property_changed(connection, adapter->path,
+				ADAPTER_INTERFACE, "Discoverable",
+				DBUS_TYPE_BOOLEAN, &discoverable);
 
 	load_drivers(adapter);
 	load_devices(adapter);
@@ -2672,20 +2657,18 @@ int adapter_stop(struct btd_adapter *adapter)
 		adapter->active_conn = NULL;
 	}
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Mode",
-					DBUS_TYPE_STRING, &mode);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Mode", DBUS_TYPE_STRING, &mode);
 
 	powered = FALSE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Powered", DBUS_TYPE_BOOLEAN, &powered);
 
 	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
 		discoverable = FALSE;
 
-		dbus_connection_emit_property_changed(connection, adapter->path,
+		emit_property_changed(connection, adapter->path,
 					ADAPTER_INTERFACE, "Discoverable",
 					DBUS_TYPE_BOOLEAN, &discoverable);
 	}
@@ -2961,7 +2944,7 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 		adapter->oor_devices = NULL;
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
+	emit_property_changed(connection, path,
 				ADAPTER_INTERFACE, "Discovering",
 				DBUS_TYPE_BOOLEAN, &discov_active);
 
@@ -3134,19 +3117,18 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 		return;
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
-						ADAPTER_INTERFACE, "Mode",
-						DBUS_TYPE_STRING, &mode);
+	emit_property_changed(connection, path, ADAPTER_INTERFACE, "Mode",
+				DBUS_TYPE_STRING, &mode);
 
 	if (powered == FALSE || adapter->scan_mode == SCAN_DISABLED) {
-		dbus_connection_emit_property_changed(connection, path,
-						ADAPTER_INTERFACE, "Powered",
-						DBUS_TYPE_BOOLEAN, &powered);
+		emit_property_changed(connection, path,
+					ADAPTER_INTERFACE, "Powered",
+					DBUS_TYPE_BOOLEAN, &powered);
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
-					ADAPTER_INTERFACE, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &discoverable);
+	emit_property_changed(connection, path,
+				ADAPTER_INTERFACE, "Discoverable",
+				DBUS_TYPE_BOOLEAN, &discoverable);
 
 	adapter_set_scan_mode(adapter, scan_mode);
 }
@@ -3248,10 +3230,29 @@ gboolean adapter_has_discov_sessions(struct btd_adapter *adapter)
 	return TRUE;
 }
 
+static void probe_driver(gpointer data, gpointer user_data)
+{
+	struct btd_adapter *adapter = data;
+	struct btd_adapter_driver *driver = user_data;
+
+	if (!adapter->up)
+		return;
+
+	driver->probe(adapter);
+}
+
 int btd_register_adapter_driver(struct btd_adapter_driver *driver)
 {
+	GSList *adapters;
+
 	adapter_drivers = g_slist_append(adapter_drivers, driver);
 
+	if (driver->probe == NULL)
+		return 0;
+
+	adapters = manager_get_adapters();
+	g_slist_foreach(adapters, probe_driver, driver);
+
 	return 0;
 }
 
@@ -3368,8 +3369,8 @@ void adapter_update_devices(struct btd_adapter *adapter)
 		devices[i] = (char *) device_get_path(dev);
 	}
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
+	emit_array_property_changed(connection, adapter->path,
 					ADAPTER_INTERFACE, "Devices",
-					DBUS_TYPE_ARRAY, &devices);
+					DBUS_TYPE_OBJECT_PATH, &devices);
 	g_free(devices);
 }
